/* $LAAS: pktIO.c,v 1.3 2005/06/14 14:46:54 matthieu Exp $ */

#if defined(POSIX)
#warning Compiling for POSIX
#elif defined(RTEMS)
#warning Compiling for RTEMS
#else
#warning Unexpected architecture
#endif

#define INNOMEDIA

/*===============================
 *======  System Includes  ======
 *===============================
 */

#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#ifdef POSIX
#include <sys/time.h>
#include <errno.h>
#endif

#ifdef RTEMS
#include <rtems.h>
#include <tmacros.h>
/* #include <uart.h> */
#endif

/*==============================
 *======  Other Includes  ======
 *==============================
 */

#include "pktIO.h"

#define PKT_STATE_WAIT 0
#define PKT_STATE_PKT   1
#define PKT_STATE_RAW  3  /* this is for returning unpacketed data */

struct pktPort {
  int fd;

  /* need read and write semaphores */

  unsigned char lastCh;
  int readState;
  int readSize;
  int readIndex;
  unsigned char readBuff[300];

  unsigned char chksum;
  int readPktSize;
  unsigned char readPkt[300];

  int writeSize;
  unsigned char writeBuff[300];

#if defined(POSIX)
#warning POSIX support incomplete
#elif defined(RTEMS)
  rtems_id readSem;
  rtems_id writeSem;
#else
#warning Unexpected Architecture
#endif
};

/*========================
 *======  Functions ======
 *========================
 */

/* ============================================================ */

void
pktAcquireReadSem(struct pktPort *pkt)
{
#if defined(POSIX)
#warning POSIX support incomplete
#elif defined(RTEMS)
   rtems_semaphore_obtain( pkt->readSem, 
			   RTEMS_WAIT, 
			   RTEMS_NO_TIMEOUT );
#else
#warning Unexpected Architecture
#endif
}

/* ============================================================ */

void
pktReleaseReadSem(struct pktPort *pkt)
{
#if defined(POSIX)
#warning POSIX support incomplete
#elif defined(RTEMS)
   rtems_semaphore_release(pkt->readSem);
#else
#warning Unexpected Architecture
#endif
}

/* ============================================================ */

static inline void
acquireWriteSem(struct pktPort *pkt)
{
#if defined(POSIX)
#warning POSIX support incomplete
#elif defined(RTEMS)
   rtems_semaphore_obtain( pkt->writeSem, 
			   RTEMS_WAIT, 
			   RTEMS_NO_TIMEOUT );
#else
#warning Unexpected Architecture
#endif
}

/* ============================================================ */

static inline void
releaseWriteSem(struct pktPort *pkt)
{
#if defined(POSIX)
#warning POSIX support incomplete
#elif defined(RTEMS)
   rtems_semaphore_release(pkt->writeSem);
#else
#warning Unexpected Architecture
#endif
}

/* ============================================================ */

/*

The function pktReadChar() performs all Esc sequence
decoding as follows:

All escape functions (STX, ETX, ESC, etc) are decoded and returned
as -EscCode. i.e. an STX is sent as 27 02 and decoded to -2 by
pktReadChar().

If there is no data, or the last character was a 27 indicating the
start of an escape sequence, a -256 is returned.

The sequence 27 0 is a special case.  It can not be used as an
escape sequence by this scheme because there is no -0.  Instead
this sequence is used to send a 27 as part of data, not an escape
sequence, so 27 0 is returned as 27.

Packet format:

STX [port] [# of data bytes] [data, incl. opcode] [chk] ETX

*/

/*
 * This is a non-blocking read.  Using select on the fd associated
 * with this stream is unreliable unless pktRead() has returned without
 * a packet.  If pktRead() returns with a packet it must be called again
 * until it returns without one or select() will not be reliable.
 */

int
pktReadChar(struct pktPort *pkt)
{
  int ch;
#ifdef POSIX
  fd_set fds;
  struct timeval timeout;
#endif

  if (pkt->readIndex >= pkt->readSize) {
#ifdef POSIX  /* this is to prevent blocking */
    timeout.tv_sec = 0;
    timeout.tv_usec = 0;
    FD_ZERO(&fds);
    FD_SET(pkt->fd, &fds);
    select(pkt->fd+1, &fds, NULL, NULL, &timeout);
    if (!FD_ISSET(pkt->fd, &fds)) {
      return(PKT_NO_DATA);
    }
#endif
    pkt->readSize = read(pkt->fd, pkt->readBuff, 299);
#if 0
    fprintf(stderr, "%s:%d - readSize=%d\n",
	    __FILE__, __LINE__, pkt->readSize);
#endif
    pkt->readIndex = 0;
  }

  if (pkt->readIndex < pkt->readSize) {
    ch = pkt->readBuff[pkt->readIndex++];
    if (pkt->lastCh == ESC) {
      if (ch==0) {
	ch=ESC;
      }
      else {
	ch=-ch;
      }

      pkt->lastCh = 0;
    }
    else {
      pkt->lastCh = ch;
      if (ch==ESC) {
	ch = pktReadChar(pkt);
      }
    }
  }
  else {
    ch = PKT_NO_DATA;
  }

  return (ch);
}

/*------------------------------------------------------------*/

int
pktRead(struct pktPort *pkt,
	unsigned char *data,
	unsigned char *port,
	unsigned char *id,
	unsigned char *opcode)
{
  unsigned char len;
  int ch;

  while (1) {
#ifdef INNOMEDIA
    do {
      ch = pktReadChar(pkt);
    } while (ch == PKT_NULL);
#else
    ch = pktReadChar(pkt);
#endif

    if (ch == PKT_NO_DATA) { /* no complete packet and no data available */
      return(0);
    }

    /*
     * I think this nested if stuff is inside out.  It might be cleaner
     * if we switch(ch) first and switch(readState) second.  I'll play
     * with it later.   -tds
     */

    if (pkt->readState == PKT_STATE_WAIT) { /* waiting for a start sequence */
      if (ch == PKT_STX) {           /* found a start */
        pkt->readState = PKT_STATE_PKT;
	pkt->readPktSize = 0;
	pkt->chksum = 0;
      }
      else if (ch >= 0) {
        /* 
	 * Unpacketed data. Return from reserved port 0.
	 * This should be improved to return a block of data
	 * instead of one byte at a time.    -tds
	 */

	len = 1;
	*port = 0;
	*opcode = 0;
	*data = ch;
	return(len);  /* might be more data available */
      }
      else {
	/* unpacketed escape code, could be XON or XOFF */
      }
    }
    else if (pkt->readState == PKT_STATE_PKT) {        /* recieving packet */
      if (ch == PKT_ETX) {           /* STOP. Need to check chkSum and len */
        pkt->readState = PKT_STATE_WAIT;
        
	if (pkt->readPktSize < 5) { /* minimum packet has 5 bytes */
	  pkt->readPktSize = 0; /* drop packet */
	  continue;
	}

        len = pkt->readPkt[3];

	if (len+5 != pkt->readPktSize) { /* length byte doesn't match */
	  pkt->readPktSize = 0; /* drop packet */
	  continue;
	}

	if (pkt->chksum != 0) { /* checksum doesn't match */
	  pkt->readPktSize = 0; /* drop packet */
	  continue;
	}

	*port   = pkt->readPkt[0];
	*id     = pkt->readPkt[1];
	*opcode = pkt->readPkt[2];
	memcpy(data, pkt->readPkt+4, len);
	return(len);  /* might be more data available */
      }
      else if (ch == PKT_STX) {
        /* misplaced start: start over */
        pkt->readState = PKT_STATE_PKT;
	pkt->readPktSize = 0;
	pkt->chksum = 0;
      }
      else if ((ch == PKT_XON) || (ch == PKT_XOFF)) {
	/* direct this information to the write functions */
      }
      else if (ch == PKT_ESC) {
	/* This might be useful for something */
      }
      else if (ch<0) {
        /* ignore unknown esc sequence */
      }
      else {
        /* part of data packet.  Add to buffer */
	if (pkt->readPktSize < 300) {
	  pkt->readPkt[pkt->readPktSize++] = ch;
	  pkt->chksum ^= (unsigned char)ch;
	}
      }
    }
  }
}

/*------------------------------------------------------------*/

/*
 * This is intended to be a reliable blocking write.
 */

int
pktWriteChar(struct pktPort *pkt, int ch)
{
  int nwrite;
  unsigned char *buff;

#ifdef POSIX
  fd_set fds;
#endif

  acquireWriteSem(pkt);

  if ((ch == PKT_STX) || (ch == PKT_CLEAR)) {
    pkt->writeSize = 0;
  }
    
  if (ch>-256) {
    if (ch<0) {
      pkt->writeBuff[pkt->writeSize++] = ESC;
      pkt->writeBuff[pkt->writeSize++] = (unsigned char)(-ch);
    }
    else {
      pkt->writeBuff[pkt->writeSize++] = (unsigned char)(ch);
      if (ch==ESC) {
	pkt->writeBuff[pkt->writeSize++] = 0;
      }
#ifdef INNOMEDIA
      else if (ch == '|') {
	pkt->writeBuff[pkt->writeSize++] = ESC;
	pkt->writeBuff[pkt->writeSize++] = (unsigned char)(-PKT_NULL);
      }
#endif
    }
  }

  if ((ch == PKT_ETX) || (ch==PKT_SEND)) {
    nwrite = 0;
    buff = pkt->writeBuff;
#if defined(POSIX)
    do {  /* this is intended to cause reliable blocking writes */

      FD_ZERO(&fds);
      FD_SET(pkt->fd, &fds);
      select(pkt->fd+1, NULL, &fds, NULL, NULL);

      errno = 0;
      pkt->writeSize -= nwrite;
      buff += nwrite;
      nwrite = write(pkt->fd, buff, pkt->writeSize);
    }
    while ((nwrite != pkt->writeSize) && (pkt->writeSize != 0) &&
	   ((errno == EAGAIN) || (errno == EINTR)));

    pkt->writeSize = 0;

    fflush(stdout);

    if (errno) {
      releaseWriteSem(pkt);
      return(-1);
    }
#elif defined(RTEMS)
    nwrite = write(pkt->fd, pkt->writeBuff, pkt->writeSize);
    pkt->writeSize = 0;
#else
#warning Unexpected Architecture
#endif
    releaseWriteSem(pkt);
    return(0);
  }
  releaseWriteSem(pkt);
  return(0);
}

/*------------------------------------------------------------*/

int
pktWrite(struct pktPort *pkt,
	 unsigned char port,
	 unsigned char id,
	 unsigned char opcode,
	 unsigned char *data,
	 int size)
{
  unsigned char len;
  unsigned char count;
  unsigned char chk;

  if ((size < 0) || (size > 255)) {
    return(-1);
  }

  len = (unsigned char)size;

  if (pktWriteChar(pkt, PKT_STX)) return(-1);

  chk = port;
  if (pktWriteChar(pkt, port)) return(-1);

  chk ^= id;
  if (pktWriteChar(pkt, id)) return(-1);

  chk ^= opcode;
  if (pktWriteChar(pkt, opcode)) return(-1);

  chk ^= len;
  if (pktWriteChar(pkt, len)) return(-1);

  for (count = 0; count<len; count++) {
    chk ^= *data;
    if (pktWriteChar(pkt, *data++)) return(-1);
  }
     
  if (pktWriteChar(pkt, chk)) return(-1);

  if (pktWriteChar(pkt, PKT_ETX)) return(-1);

  return (len);
}

/*------------------------------------------------------------*/

int pktGetFd(struct pktPort *pkt) {
  return(pkt->fd);
}

/*------------------------------------------------------------*/

void
pktAddChar(unsigned char *buffer, int *pos, unsigned char data)
{
  buffer[(*pos)++] = data;
}

/*------------------------------------------------------------*/

void
pktAddShort(unsigned char *buffer, int *pos, unsigned short data)
{
  int pos2;

  pos2 = *pos;

  buffer[pos2  ] = (unsigned char)((data >> 8) & 0xFF);
  buffer[pos2+1] = (unsigned char)((data     ) & 0xFF);

  *pos = pos2+2;
}

/*------------------------------------------------------------*/

void
pktAddLong(unsigned char *buffer, int *pos, unsigned long data)
{
#if 0 /* doesn't work */
  buffer[*pos] = (unsigned char)((data >> 24) & 0xFF); (*pos)++;
  buffer[*pos] = (unsigned char)((data >> 16) & 0xFF); (*pos)++;
  buffer[*pos] = (unsigned char)((data >>  8) & 0xFF); (*pos)++;
  buffer[*pos] = (unsigned char)((data      ) & 0xFF); (*pos)++;
#else
  int pos2;

  pos2 = *pos;

  buffer[pos2  ] = (unsigned char)((data >> 24) & 0xFF);
  buffer[pos2+1] = (unsigned char)((data >> 16) & 0xFF);
  buffer[pos2+2] = (unsigned char)((data >>  8) & 0xFF);
  buffer[pos2+3] = (unsigned char)((data      ) & 0xFF);

  *pos = pos2 + 4;
#endif
}

/*------------------------------------------------------------*/

void
pktAdd(unsigned char *buffer, int *pos, unsigned char *data, int nAdd)
{
  memcpy(&buffer[*pos], data, nAdd);
  *pos += nAdd;
}

/*------------------------------------------------------------*/

unsigned char
pktGetChar(unsigned char *buffer, int *pos)
{
  return(buffer[(*pos)++]);
}

/*------------------------------------------------------------*/

unsigned short
pktGetShort(unsigned char *buffer, int *pos)
{
  /*
   * Stewart reports that casting of unsigned char to
   * unsigned short is broken.  Using an intermediate
   * unsigned int is a workaround.
   *
   * Additionally, the evalutation order of *pos++ isn't
   * defined and GCC doesn't do what I had assumed.
   */
#if 0
  return((unsigned long)((buffer[(*pos)++] << 8)) |
	 (unsigned long)((buffer[(*pos)++])));
#else
  unsigned int ret0, ret1;
  int pos2;

  pos2 = *pos;
  ret1 = buffer[pos2] << 8;
  ret0 = buffer[pos2+1];
  *pos = pos2+2;
  return(ret1 | ret0);
#endif
}

/*------------------------------------------------------------*/

unsigned long
pktGetLong(unsigned char *buffer, int *pos)
{
  unsigned long ret0, ret1, ret2, ret3;
  int pos2;

  pos2 = *pos;

  ret3 = buffer[pos2  ] << 24;
  ret2 = buffer[pos2+1] << 16;
  ret1 = buffer[pos2+2] <<  8;
  ret0 = buffer[pos2+3];

  *pos = pos2+4;

  return(ret3 | ret2 | ret1 | ret0);
}

/*------------------------------------------------------------*/

void
pktGet(unsigned char *buffer, int *pos, unsigned char *data, int nGet)
{
  memcpy(data, &buffer[*pos], nGet);
  *pos += nGet;
}

/*------------------------------------------------------------*/

struct pktPort *
pktInitPort(int fd) {
  struct pktPort *pkt;

  pkt = malloc(sizeof(*pkt));
  
  pkt->fd = fd;
  pkt->lastCh = 0;
  pkt->readState = PKT_STATE_WAIT;
  pkt->readSize = 0;
  pkt->readIndex = 0;
  pkt->chksum = 0;
  pkt->readPktSize = 0;
  pkt->writeSize = 0;

#if defined(POSIX)
#warning POSIX support incomplete
#elif defined(RTEMS)
  {
    rtems_name	        semName;
    
    semName = rtems_build_name( 'P', 'K', 'R', fd + '0' );
    rtems_semaphore_create( semName,
			    1,		/* initial count */
			    (RTEMS_PRIORITY |
			     RTEMS_BINARY_SEMAPHORE |
			     RTEMS_INHERIT_PRIORITY |
			     RTEMS_NO_PRIORITY_CEILING |
			     RTEMS_LOCAL),
			    1,		/* priority ceiling? */
			    &pkt->readSem);

    semName = rtems_build_name( 'P', 'K', 'W', fd + '0' );
    rtems_semaphore_create( semName,
			    1,		/* initial count */
			    (RTEMS_PRIORITY |
			     RTEMS_BINARY_SEMAPHORE |
			     RTEMS_INHERIT_PRIORITY |
			     RTEMS_NO_PRIORITY_CEILING |
			     RTEMS_LOCAL),
			    1,		/* priority ceiling? */
			    &pkt->writeSem);
  }
#else
#warning Unexpected Architecture
#endif

  return(pkt);
}
