#include <unistd.h>
#include <stdlib.h> 
#include <stdio.h> 
#include <string.h>
#include <termios.h>
#include <sys/time.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <signal.h>
#include <ctype.h>
#include <time.h>

#include <rFlex.h>
#include "serial.h"

#define LONG_PAUSE 50
#define SHORT_PAUSE 0

#ifndef MAX
#define MAX(x, y) (( (x) > (y) ) ? (x) : (y))
#endif
#ifndef MIN
#define MIN(x, y) (( (x) < (y) ) ? (x) : (y))
#endif

static char filename[256];
static char port[256];
static char nBaudStr[256];
static speed_t nBaud;
static char uBaudStr[256];
static speed_t uBaud;
static int driveOn = 0;
static int jx, jy;
static unsigned char jb;
static int tVel, rVel;

static int idSetDir[4];
static int idJstk;
static int idGetSys[2];

static int dioTS;
static int dioCount;
static int dioOn;
static int dioState = 0;

/*------------------------------------------------------------*/
 
/*
* This code sets up a terminal for single keystroke I/O
* and returns it to its original state when needed.
*/

static struct termios save_termios;
static int ttysavefd = -1;
static enum {RESET, CBREAK} ttystate = RESET;
int tty_reset(int fd);
static void sig_catch(int signo);

/*------------------------------------------------------------*/

int tty_cbreak(int fd) { /* put terminal in cbreak mode */
  struct termios buf;
  
  if (tcgetattr(fd, &save_termios) < 0)
    return (-1);

  buf=save_termios;
  buf.c_lflag &= ~(ECHO | ICANON); /* echo off, cononacal mode off */
  buf.c_cc[VMIN] = 0; /* Case B: 1 byte at a time, no timer */
  buf.c_cc[VTIME] = 0;

  if (tcsetattr(fd, TCSAFLUSH, &buf) < 0)
    return (-1);
  ttystate = CBREAK;
  ttysavefd=fd;
  return (0);
}

/*------------------------------------------------------------*/

int tty_reset(int fd) { /* restore terminal's mode */
  if (ttystate != CBREAK)
    return (0);

  if (tcsetattr(fd, TCSAFLUSH, &save_termios) <0)
    return (-1);
  ttystate=RESET;
  return (0);
}

/*------------------------------------------------------------*/

void
ttyExit(void)
{
  fprintf(stderr, "Resetting tty mode... ");
  
  if (tty_reset(STDIN_FILENO)<0) {
    fprintf(stderr, "tty_reset() error\n");
  }
  else {
    fprintf(stderr, "done.\n");
  }
  
  fprintf(stderr, "Exiting.\n");

  exit(0);
}

static void sig_catch(int signo) {
  ttyExit();
}

/*------------------------------------------------------------*/


/* ============================================================ */
/* ============================================================ */

static int
sysLcdDumpRpl(struct reflex *reflex, int rv, unsigned char id, int ts,
	      unsigned char row, unsigned char nbytes,
	      unsigned char *data)
{
  static int count = 0;
  static FILE *fp = NULL;
  unsigned char fileName[256];

  int ii, i2, col;

  fprintf(stderr, "%s:%d:%s() - row:%d nbytes:%d count:%d\n",
	  __FILE__, __LINE__, __FUNCTION__, row, nbytes, count);

  if (!fp) {
    sprintf(fileName, "reflexScreenDump-%04d.pgm", count++);
    fp = fopen(fileName, "w");

    if (!fp) {
      return(-1);
    }

    fprintf(fp, "P2 320 240 1\n");
  }

  col = 0;

  for (ii=0; ii<nbytes; ii++) {
    for (i2 = 0; i2<8; i2++) {
      if (col++ > 20) {
	fprintf(fp, "\n  ");
	col = 0;
      }
      fprintf(fp, " %d", (data[ii] & (1<<i2)) == 0);
    }
  }

  fprintf(fp, "\n");

  if (row==239) {
    fclose(fp);
    fp = NULL;
  }

  return(0);
}

/* ------------------------------------------------------------ */

static int
sysStatusRpl(struct reflex *reflex, int rv, unsigned char id, int ts,
	     int volts, unsigned char brake)
{

  fprintf(stderr, "SYS_STATUS: volts=%d brake=%d\n", volts, brake);
  return(0);  
}

/* ------------------------------------------------------------ */

static int
sysUnknownRpl(struct reflex *reflex, unsigned char id,
	      unsigned char size, unsigned char *buffer,
	      unsigned char opcode)
{
  int ii;

  fprintf(stderr, "%s:%6d:%s() - id=%d opcode=%d : ",
	  __FILE__, __LINE__, __FUNCTION__, id, opcode);

  for (ii=0; ii<size; ii++) {
    fprintf(stderr, " 0x%02X", ((unsigned)buffer[ii]));
  }
  fprintf(stderr, "\n");

  return(0);  
}

/* ============================================================ */
/* ============================================================ */

/* motion control */

static int
motAxisSystemRpt (struct reflex *reflex,
		  unsigned char id, int ts,
		  unsigned char axis,
		  int pos, int vel, int acc, int trq)
{

  if (axis == 0) {
    tVel = vel;
  }
  else if (axis == 1) {
    rVel = vel;
  }
  
  fprintf(stderr, "%s:%6d:%s()\n", __FILE__, __LINE__, __FUNCTION__);
  fprintf(stderr, "\tid        = %d\n", id);
  fprintf(stderr, "\ttimeStamp = %d\n", ts);
  fprintf(stderr, "\taxis      = %d\n", axis);
  fprintf(stderr, "\tpos       = %d\n", pos);
  fprintf(stderr, "\tvel       = %d\n", vel);
  fprintf(stderr, "\tacc       = %d\n", acc);
  fprintf(stderr, "\ttrq       = %d\n", trq);
  
  return(0);
}

/* ------------------------------------------------------------ */

static int
motAxisGetSystemRpl (struct reflex *reflex, int rv,
		     unsigned char id, int ts,
		     unsigned char axis,
		     int pos, int vel, int acc, int trq)
{  

  if (axis == 0) {
    tVel = vel;
  }
  else if (axis == 1) {
    rVel = vel;
  }
  
  if (id == 255) {
    fprintf(stderr, "%s:%6d:%s()\n\tretval = %d\n",
	    __FILE__, __LINE__, __FUNCTION__, rv);
    fprintf(stderr, "\tid        = %d\n", id);
    fprintf(stderr, "\ttimeStamp = %d\n", ts);
    fprintf(stderr, "\taxis      = %d\n", axis);
    fprintf(stderr, "\tpos       = %d\n", pos);
    fprintf(stderr, "\tvel       = %d\n", vel);
    fprintf(stderr, "\tacc       = %d\n", acc);
    fprintf(stderr, "\ttrq       = %d\n", trq);
  }
  
  idGetSys[axis] = id;
  
  return(0);
}

/* ------------------------------------------------------------ */

static int
motAxisGetModelRpl (struct reflex *reflex, int rv,
		    unsigned char id, int ts,
		    unsigned char axis,
		    int pos, int vel, int acc, int trq)
{  

  if (axis == 0) {
    tVel = vel;
  }
  else if (axis == 1) {
    rVel = vel;
  }
  
  if (id == 255) {
    fprintf(stderr, "%s:%6d:%s()\n\tretval = %d\n",
	    __FILE__, __LINE__, __FUNCTION__, rv);
    fprintf(stderr, "\tid        = %d\n", id);
    fprintf(stderr, "\ttimeStamp = %d\n", ts);
    fprintf(stderr, "\taxis      = %d\n", axis);
    fprintf(stderr, "\tpos       = %d\n", pos);
    fprintf(stderr, "\tvel       = %d\n", vel);
    fprintf(stderr, "\tacc       = %d\n", acc);
    fprintf(stderr, "\ttrq       = %d\n", trq);
  }
  
  idGetSys[axis] = id;
  
  return(0);
}

/* ------------------------------------------------------------ */

static int
motAxisGetTargetRpl (struct reflex *reflex, int rv,
		     unsigned char id, int ts,
		     unsigned char axis,
		     int pos, int vel, int acc, int trq)
{  

  if (axis == 0) {
    tVel = vel;
  }
  else if (axis == 1) {
    rVel = vel;
  }
  
  if (id == 255) {
    fprintf(stderr, "%s:%6d:%s()\n\tretval = %d\n",
	    __FILE__, __LINE__, __FUNCTION__, rv);
    fprintf(stderr, "\tid        = %d\n", id);
    fprintf(stderr, "\ttimeStamp = %d\n", ts);
    fprintf(stderr, "\taxis      = %d\n", axis);
    fprintf(stderr, "\tpos       = %d\n", pos);
    fprintf(stderr, "\tvel       = %d\n", vel);
    fprintf(stderr, "\tacc       = %d\n", acc);
    fprintf(stderr, "\ttrq       = %d\n", trq);
  }
  
  idGetSys[axis] = id;
  
  return(0);
}

/* ------------------------------------------------------------ */

static int
motAxisSetDirRpl (struct reflex *reflex, int rv,
		  unsigned char id, int ts,
		  unsigned char axis)
{

#if 0
  fprintf(stderr, "MOT_AXIS_SET_DIR:  retval=%d id=%d ts=%d axis=%d\n",
	  rv, id, ts, axis);
#endif
  idSetDir[axis] = id;
  
  return(0);
}

/* ------------------------------------------------------------ */

static int
motSetDefaultsRpl (struct reflex *reflex, int rv,
		   unsigned char id, int ts)
{
  fprintf(stderr, "%s:%6d:%s()\n", __FILE__, __LINE__, __FUNCTION__);
  return(0);
}

/* ------------------------------------------------------------ */

static int
motBrakeSetRpl(struct reflex *reflex, int rv,
	       unsigned char id, int ts)
{
  fprintf(stderr, "%s:%6d:%s()\n", __FILE__, __LINE__, __FUNCTION__);
  return(0);
}

/* ------------------------------------------------------------ */

static int
motBrakeReleaseRpl(struct reflex *reflex, int rv,
		   unsigned char id, int ts)
{

  fprintf(stderr, "%s:%6d:%s() - retval = %d\n",
	  __FILE__, __LINE__, __FUNCTION__, rv);
  return(0);  
}

/* ------------------------------------------------------------ */

static int
motSetSystemRptRpl (struct reflex *reflex, int rv,
		    unsigned char id, int ts)
{
  fprintf(stderr, "%s:%6d:%s()\n", __FILE__, __LINE__, __FUNCTION__);
  return(0);
}

/* ------------------------------------------------------------ */

static int
motUnknownRpl(struct reflex *reflex, unsigned char id,
	      unsigned char size,
	      unsigned char *buffer, unsigned char opcode)
{
  int ii;

  fprintf(stderr, "%s:%6d:%s() - id=%d opcode=%d : ",
	  __FILE__, __LINE__, __FUNCTION__, id, opcode);

  for (ii=0; ii<size; ii++) {
    fprintf(stderr, " 0x%02X", ((unsigned)buffer[ii]));
  }
  fprintf(stderr, "\n");

  return(0);
}

/* ============================================================ */
/* ============================================================ */

static int
jstkGetStateRpl(struct reflex *reflex, int rv, unsigned char id, int ts,
		int x, int y, unsigned char buttons)
{

  jx = x;
  jy = y;
  jb = buttons;
  
  if (id == 255) {
    fprintf(stderr, "JSTK_GET_STATE:\n");
    fprintf(stderr, "\tid        = %d\n", id);
    fprintf(stderr, "\ttimeStamp = %d\n", ts);
    fprintf(stderr, "\tx         = %d\n", jx);
    fprintf(stderr, "\ty         = %d\n", jy);
    fprintf(stderr, "\tbuttons   = %d\n", jb);
  }
  
  idJstk = id;

  return(0);
}

/* ------------------------------------------------------------ */

static int
jstkUnknownRpl(struct reflex *reflex, unsigned char id, unsigned char size,
	       unsigned char *buffer, unsigned char opcode)
{
  int ii;

  fprintf(stderr, "%s:%6d:%s() - id=%d opcode=%d : ",
	  __FILE__, __LINE__, __FUNCTION__, id, opcode);

  for (ii=0; ii<size; ii++) {
    fprintf(stderr, " 0x%02X", ((unsigned)buffer[ii]));
  }
  fprintf(stderr, "\n");

  return(0);
}

/* ============================================================ */
/* ============================================================ */

static int
sonarRptRpl(struct reflex *reflex, int rv, unsigned char id, int ts)
{
  fprintf(stderr, "SONAR_RUN: retval = %d\n", rv);
  return(0);
}

/* ------------------------------------------------------------ */

int
sonarGetUpdateRpl (struct reflex *reflex, int rv, unsigned char id, int ts,
		   int count, unsigned char *sonarId,
		   unsigned short *sonarValue)
{
  int ii;

  fprintf(stderr, "SONAR_GET_UPDATE: count=%d timeStamp = %d\n",
	  count, ts);

  for (ii=0; ii<count; ii++) {
    fprintf(stderr, "   SONAR %02X: %5d\n",
	    sonarId[ii], sonarValue[ii]);
  }

  fprintf(stderr, "\n");

  return(0);
}

/* ------------------------------------------------------------ */

int
sonarRpt (struct reflex *reflex, unsigned char id, int ts,
	  int count, unsigned char *sonarId, unsigned short *sonarValue)
{
  int ii;

  fprintf(stderr, "SONAR_REPORT: count=%d timeStamp = %d\n",
	  count, ts);

  for (ii=0; ii<count; ii++) {
    fprintf(stderr, "   SONAR %02X: %5d\n",
	    sonarId[ii], sonarValue[ii]);
  }
  
  fprintf(stderr, "\n");

  return(0);
}

/* ------------------------------------------------------------ */

int
sonarUnknownRpl(struct reflex *reflex, unsigned char id, unsigned char size,
		unsigned char *buffer, unsigned char opcode)
{
  int ii;

  fprintf(stderr, "%s:%6d:%s() - id=%d opcode=%d : ",
	  __FILE__, __LINE__, __FUNCTION__, id, opcode);
  for (ii=0; ii<size; ii++) {
    fprintf(stderr, " 0x%02X", ((unsigned)buffer[ii]));
  }
  fprintf(stderr, "\n");

  return(0);
}

/* ============================================================ */
/* ============================================================ */

int
dioUpdateRpl (struct reflex *reflex, int rv, unsigned char id, int ts,
		   int count)
{

  fprintf(stderr, "%s() - rv=%d count=%d timeStamp = %d\n",
	  __FUNCTION__, rv, count, ts);

  if (ts<dioTS) {
    dioTS = 0; /* time stamp wrapped or reset, next req will get update */
  }
  else if (dioCount == count) {
    dioTS = ts;
  }

  return(0);
}

/* ------------------------------------------------------------ */

int
dioUpdate (struct reflex *reflex, unsigned char id, int ts,
	   unsigned char netId, unsigned char offset, unsigned char state)
{

  fprintf(stderr, "%s:%6d:%s() - netId:0x%02X offset:%d state:0x%02X\n",
	  __FILE__, __LINE__, __FUNCTION__,
	  netId, offset, state);

  dioCount++;

  return(0);
}

/* ------------------------------------------------------------ */

static int
dioRptRpl(struct reflex *reflex, int rv, unsigned char id, int ts)
{
  fprintf(stderr, "%s() - retval = %d\n", __FUNCTION__, rv);
  return(0);
}

/* ------------------------------------------------------------ */

int
dioRpt (struct reflex *reflex, unsigned char id, int ts,
	unsigned char netId, unsigned char offset, unsigned char state)
{

  fprintf(stderr, "%s:%6d:%s() - netId:0x%02X offset:%d state:0x%02X\n",
	  __FILE__, __LINE__, __FUNCTION__,
	  netId, offset, state);

  return(0);
}

/* ------------------------------------------------------------ */

int
dioUnknownRpl(struct reflex *reflex, unsigned char id, unsigned char size,
	      unsigned char *buffer, unsigned char opcode)
{
  int ii;

  fprintf(stderr, "%s:%6d:%s() - id=%d opcode=%d : ",
	  __FILE__, __LINE__, __FUNCTION__, id, opcode);
  for (ii=0; ii<size; ii++) {
    fprintf(stderr, " 0x%02X", ((unsigned)buffer[ii]));
  }
  fprintf(stderr, "\n");

  return(0);
}

/* ============================================================ */
/* ============================================================ */

/*
 * findPrompt() is hokey because it isn't ready to handle arbitrary
 * values for size and because it assumes that prompt and newDat
 * are ASCII-Z strings so it won't work properly on a binary
 * stream.  If anyone ever cares it could be reimplemented to
 * handle any size<X (for some X) and binary streams by doing 
 * resonable checks on size and not using strxxx() functions.
 * One good reason for doing this is that a '\0' could show up
 * randomly (and hard to repeat!) from power cycles or baud
 * changes on the device.
 *
 * It also should be passed a handle to its static data so that
 * more than one stream at a time can be checked.  local static
 * is a bogus hack!    -TDS
 */

int
findPrompt(int size, unsigned char *newDat, unsigned char *prompt)
{
  static unsigned char promptBuffer[33];
  static int bufferSize = 0;
  static int promptPos = 0;
  unsigned char *ptr;
  int ii;

  /* because thats the way it is for now */
  if (size != 1) {
    fprintf(stderr, 
	    "findPrompt() isn't ready to assure reliability if size!=1. "
	    "size = %d\n", size);
    return(0);
  }

  /* make space in the head for the new data */

  ii = bufferSize + size - 32;
  if (ii>0) {
#ifdef SUNOS
    memcpy (promptBuffer, promptBuffer+ii, strlen(promptBuffer)-ii);
#else
    memmove(promptBuffer, promptBuffer+ii, strlen(promptBuffer)-ii);
#endif
    bufferSize -= ii;
    promptPos  -= ii;
  }

  /* ensure the buffer is NULL terminated and append the new data */

  promptBuffer[bufferSize] = 0;
  strncat(promptBuffer, newDat, size);
  bufferSize += size;
  promptBuffer[bufferSize] = 0;

  /* look for the prompt and update promptPos */

  ptr = strstr(promptBuffer+promptPos, prompt);

  if (ptr) {
    promptPos = ptr - promptBuffer + strlen(prompt);
    return(1);
  }
  else {
    promptPos = bufferSize - strlen(prompt) + 1;
    return(0);
  }
}

/* ============================================================ */

/* Port 0 is unpacketed data */

static int
defaultHandlePort(struct reflex *reflex, int size, unsigned char *buffer, 
		  unsigned char port, unsigned char id, unsigned char opcode)
{
  static int state = 1;  /* all this static stuff is A Bad Thing (tm) */
  static FILE *app;
  static int count;
  static time_t start;
  static char lbuffer[1024];
  time_t stop;
  int fd;
  struct pktPort *pkt;

  if (opcode==1) {
    fflush(stdout);
    fprintf(stderr, "[%s:%6d:%s() - size = %d]\n",
	    __FILE__, __LINE__, __FUNCTION__, size);
    return(0);
  }
  else if (opcode==2) {
    fflush(stdout);
    fprintf(stderr, "[%s:%6d:%s() - unknown port = %d]\n",
	    __FILE__, __LINE__, __FUNCTION__, port);
    return(0);
  }

  pkt = reflex->pktPort;

  fd = pktGetFd(pkt);

  if ((size>0) && (opcode==0)) {
    buffer[size] = 0;
    fprintf(stderr, "%s", buffer);
  }
  else if ((size>0) && (opcode==1)) {
    int ii;

    fprintf(stderr, "%s:%6d:%s() - unknown port=%d :",
	    __FILE__, __LINE__, __FUNCTION__, id);
    for (ii=0; ii<size; ii++) {
      fprintf(stderr, " 0x%02X", ((unsigned)buffer[ii]));
    }
    fprintf(stderr, "\n");
    return(0);
  }

  switch (state) {
  case 1:
    if (size == -1) {
      /* open file to upload */
      fprintf(stderr, "Opening %s for upload\n", filename);
      app = fopen(filename, "r");
      if (!app) {
	fprintf(stderr, "unable to open application file\n");
	break;
      }
    
      fprintf(stderr, "Starting application upload...\n");
    
      count = 0;
      start = time(NULL);
      state = 2;
#if 0
      reflex->motion->setDefaultsReq(reflex, id);
      driveOn = 0;
#endif
    }
    else if (size == -2) {
      if (driveOn) {
	fprintf(stderr, "%s:%6d:%s() - switch drive off\n",
		__FILE__, __LINE__, __FUNCTION__);
	reflex->motion->setDefaultsReq(reflex, id);
	reflex->sonar->rptReq(reflex, id, 0, 0, 0, 0);
	driveOn = 0;
      }
      else {
        fprintf(stderr, "%s:%6d:%s() - setting defaults\n",
                __FILE__, __LINE__, __FUNCTION__);
        reflex->motion->setDefaultsReq(reflex, id);
	reflex->sonar->rptReq(reflex, id, 30000, 0, 0, 2);
	driveOn = 1;
      }
    }
    break;

  case 2:
    if (size<0) {
      break;
    }
    if (size==0) {
      strcpy(lbuffer, "\r");
      write(fd, lbuffer, strlen(lbuffer));
      if (count++>10) {
	fprintf(stderr, ".");
	count = 0;
      }
    }
    else {
      if (size>0 && findPrompt(size, buffer, "dbg>")) {
	sprintf(lbuffer, "baud %s\n", uBaudStr);
	write(fd, lbuffer, strlen(lbuffer));
	usleep(LONG_PAUSE * 1000);
	state = 3;
	break;
      }
      count = 0;
    }
    break;

  case 3:
    if (0 || (size>0 && findPrompt(size, buffer, uBaudStr))) {
      setBaudRate(fd, uBaud);
      usleep(LONG_PAUSE * 1000);
      state = 4;
    }
    break;

  case 4:
    if (size==0) {
      strcpy(lbuffer, "\r");
      write(fd, lbuffer, strlen(lbuffer));
      usleep(LONG_PAUSE * 1000); /* stall for 100ms */
      strcpy(lbuffer, "load\r");
      write(fd, lbuffer, strlen(lbuffer));
      usleep(LONG_PAUSE * 1000); /* stall for 100ms */
      state = 5;
    }
    break;

  case 5:
    if (0 || (size>0 && findPrompt(size, buffer, ":"))) {
      /*
       * This won't work properly if an S-rec is larger than 1023 bytes.
       * I think that an S-rec has an 8 bit len field so the record should
       * never be much more than 512 bytes.   -TDS
       */

      if (!fgets(lbuffer, 1023, app)) {
	strcpy(lbuffer, "\r");
	write(fd, lbuffer, strlen(lbuffer));
	usleep(LONG_PAUSE * 1000);
	state = 6;
	break;
      }

      if (lbuffer[strlen(lbuffer)-2] == 13) {
	lbuffer[strlen(lbuffer)-2] = '\n';
	lbuffer[strlen(lbuffer)-1] = 0;
      }
      write(fd, lbuffer, strlen(lbuffer));
      fflush(stdout);
      usleep(SHORT_PAUSE * 1000); /* stall for 30ms */
    }
    break;

  case 6:
    if (size>0 && findPrompt(size, buffer, "dbg>")) {
      sprintf(lbuffer, "baud %s\r", nBaudStr);
      write(fd, lbuffer, strlen(lbuffer));
      usleep(LONG_PAUSE * 1000);
      state = 7;
    }
    break;

  case 7:
    if (size>0 && findPrompt(size, buffer, nBaudStr)) {
      setBaudRate(fd, nBaud);
      strcpy(lbuffer, "\r");
      write(fd, lbuffer, strlen(lbuffer));
      usleep(LONG_PAUSE * 1000);
      strcpy(lbuffer, "run\r");
      write(fd, lbuffer, strlen(lbuffer));
      usleep(LONG_PAUSE * 1000);
      state = 8;
    }
    break;

  case 8:
    if (size>0 && findPrompt(size, buffer, "run")) {
      stop = time(NULL);
      fclose(app);
      fprintf(stderr, "\nelapsed time = %ld seconds\n", stop-start);
      state = 1;
    }
    break;

  default:
    fprintf(stderr, "*** unexpected state %d. setting state = 1\n", state);
    state = 1;
  }

  {
    static unsigned char id = 0;
    static unsigned char drvcount = 0;
    static time_t drvtime1 = 0;
    static time_t drvtime2 = 0;

    if (driveOn &&
	((!size) || (idGetSys[1] == id))
	) {
      int trq, acc, vel;
      unsigned char dir;
      
      if (idSetDir[0]!=id)
	fprintf(stderr, "\nMOT_AXIS_SET_DIR 0:    id = %d", id);

      if (idSetDir[1]!=id)
	fprintf(stderr, "\nMOT_AXIS_SET_DIR 1:    id = %d", id);
#if 0
      if (idSetDir[2]!=id)
	fprintf(stderr, "\nMOT_AXIS_SET_DIR 2:    id = %d", id);

      if (idSetDir[3]!=id)
	fprintf(stderr, "\nMOT_AXIS_SET_DIR 3:    id = %d", id);
#endif      
      if (idJstk!=id)
	fprintf(stderr, "\nJSTK_GET_STATE:        id = %d", id);
      
      if (idGetSys[0]!=id)
	fprintf(stderr, "\nMOT_AXIS_GET_SYSTEM 0: id = %d", id);

      if (idGetSys[1]!=id)
	fprintf(stderr, "\nMOT_AXIS_GET_SYSTEM 1: id = %d", id);
      
      id++;
      if (id > 254) id = 0;

#if 0
      fprintf(stderr, "\njb=%d                 id=%4d\n", jb, id);
#else
      fprintf(stderr, "*");
#endif
      
      if (jb==1) {
	vel = jy * 2000;
	acc = MAX(200000, abs(vel - tVel)*3);
	dir = (vel<0 ? 0 : 1);
	vel = abs(vel);
	trq = 30000;
	
	reflex->motion->setDirReq(reflex, id, 0, vel,  acc,  trq, dir);
	
	vel = jx * (-1000);
	acc = MAX(100000, abs(vel - rVel)*4);
	dir = (vel<0 ? 0 : 1);
	vel = abs(vel);
	trq = 35000;
	
	reflex->motion->setDirReq(reflex, id, 1, vel,  acc,  trq, dir);

#if 0
	vel = 0;
	dir = 0;
	acc = 500000;
	trq = 160000;
	
	reflex->motion->setDirReq(reflex, id, 2, vel,  acc,  trq, dir);
	reflex->motion->setDirReq(reflex, id, 3, vel,  acc,  trq, dir);
#endif
      }
      else {
	vel = 0;
	dir = 0;
	acc = 1000000;
	trq = 35000;
	
	reflex->motion->setDirReq(reflex, id, 0, vel,  acc,  trq, dir);
	
	trq = 1000;
	
	reflex->motion->setDirReq(reflex, id, 1, vel,  acc,  trq, dir);
#if 0
	vel = 0;
	dir = 0;
	acc = 500000;
	trq = 160000;
	
	reflex->motion->setDirReq(reflex, id, 2, vel,  acc,  trq, dir);
	reflex->motion->setDirReq(reflex, id, 3, vel,  acc,  trq, dir);
#endif
      }
      
#if 0
      fprintf(stderr, "tVel = %10d rVel=%10d jx=%4d jy=%4d jb=%d\n",
	      tVel, rVel, jx, jy, jb);
#endif
      reflex->joystick->stateReq(reflex, id);
      reflex->motion->getSystemReq(reflex, id, 0);
      reflex->motion->getSystemReq(reflex, id, 1);

      drvtime2 = time(NULL);
      if (drvtime2-drvtime1>10) {
	fprintf(stderr, "\ndrive updates = %d :",
		(unsigned)(unsigned char)(id-drvcount));
	drvcount = id;
	drvtime1 = drvtime2;
      }
    }
  }
  return(0);
}

/* ============================================================ */

void
handleStdin(struct reflex *reflex, int size, unsigned char *buffer) {
  int ii;
  int fd;
  unsigned char id;
  struct pktPort *pkt;

  pkt = reflex->pktPort;

  id = 255;

  
  fd = pktGetFd(pkt);

  for (ii=0; ii<size; ii++) {
    if (buffer[ii] == 'W'-'A'+1) {
      fprintf(stderr, "requesting system state\n");
      reflex->motion->getSystemReq(reflex, id, 0);
    }
    else if (buffer[ii] == 'I'-'A'+1) {
      fprintf(stderr, "starting system reports\n");
      reflex->motion->setSystemRptReq(reflex, id, 100000, 0x3);
    }
    else if (buffer[ii] == 'O'-'A'+1) {
      fprintf(stderr, "stopping system reports\n");
      reflex->motion->setSystemRptReq(reflex, id, 100000, 0x0);
    }
    else if (buffer[ii] == 'P'-'A'+1) {
      fprintf(stderr, "requesting system status\n");
      reflex->system->statusReq(reflex, id);
    }
    else if (buffer[ii] == 'N'-'A'+1) {
      fprintf(stderr, "releasing brake\n");
      reflex->motion->brakeReleaseReq(reflex, id);
    }
#if 0
    else if (buffer[ii] == 'K'-'A'+1) {
      fprintf(stderr, "setting brake\n");
      reflex->motion->brakeSetReq(reflex, id);
    }
#else
    else if (buffer[ii] == 'K'-'A'+1) {
      fprintf(stderr, "switching lights\n");
      reflex->dio->setReq(reflex, id, 0x28, 0, (dioState++)<<4, 0x30);
    }
#endif
    else if (buffer[ii] == 'E'-'A'+1) {
      fprintf(stderr, "requesting dio update: dioTS=%d\n", dioTS);
      dioCount = 0;
      reflex->dio->updateReq(reflex, id, dioTS);
    }
    else if (buffer[ii] == 'R'-'A'+1) {
      if (dioOn) {
	dioOn = 0;
	fprintf(stderr, "stopping dio reports\n");
	reflex->dio->rptReq(reflex, id, dioOn);
      }
      else {
	dioOn = 10000;
	fprintf(stderr, "starting dio reports\n");
	reflex->dio->rptReq(reflex, id, dioOn);
      }
    }
    else if (buffer[ii] == 'T'-'A'+1) {  
      fprintf(stderr, "requesting jstk state\n");
      reflex->joystick->stateReq(reflex, id);
    }
    else if (buffer[ii] == 'B'-'A'+1) { /* toggle drive function */
      defaultHandlePort(reflex, -2, buffer, 0, 255, 255);
    }
    else if (buffer[ii] == 'V'-'A'+1) {
      reflex->system->lcdDumpReq(reflex, id);
    }
    else if (buffer[ii] == 'U'-'A'+1) {
      defaultHandlePort(reflex, -1, buffer, 0, 255, 255); /* start upload */
    }
    else if (!isprint(buffer[ii]) &&
	     (buffer[ii]!=0x08) &&
	     (buffer[ii]!=0x09) &&
	     (buffer[ii]!=0x0A) &&
	     (buffer[ii]!=0x7F)) {
      fprintf(stderr, "[0x%02X]", buffer[ii]);
    }
    else {
      write(fd, &(buffer[ii]), 1);
    }
  }
}

/* ============================================================ */
/* ============================================================ */

static void readLoop(struct reflex *reflex, int fd)
{
  char buffer[256];
  int nread;
  fd_set fds;
  struct timeval timeout;

  do {
    timeout.tv_sec = 0;
    timeout.tv_usec = 100000;
    FD_ZERO(&fds);
    FD_SET(STDIN_FILENO, &fds);
    FD_SET(fd, &fds);
    select(fd+1, &fds, NULL, NULL, &timeout);

    if (FD_ISSET(fd, &fds)) {
      do {
	nread = reflex->read(reflex, 5);
	defaultHandlePort(reflex, -3, buffer, 0, 255, 255);
      } while (nread != 0);
    }
    else if (FD_ISSET(STDIN_FILENO, &fds)) { /* copy from stdin to port */
      nread = read(STDIN_FILENO, buffer, 255);
      buffer[nread] = 0;
      handleStdin(reflex, nread, buffer);
    }
    else {
      defaultHandlePort(reflex, 0, buffer, 0, 255, 255);
    }
  } while (1);
}

/* ============================================================ */

int main(int argc, char *argv[]) {
  int fd;
  struct reflex *reflex;

  strcpy(filename, "init.exe");
  strcpy(port, "/dev/cua1");
  strcpy(nBaudStr, "38400");
  strcpy(uBaudStr, "38400");
  
  if (argc > 1) {
    strcpy(port, argv[1]);
  }
  fprintf(stderr, "port = %s\n", port);

  if (argc > 2) {
    strcpy(filename, argv[2]);
  }
  fprintf(stderr, "filename = %s\n", filename);

  if (argc > 3) {
    strcpy(nBaudStr, argv[3]);
  }
  fprintf(stderr, "normal baud = %s\n", nBaudStr);
    
  if (argc > 4) {
    strcpy(uBaudStr, argv[4]);
  }
  else if (argc > 3) {
    strcpy(uBaudStr, argv[3]);
  }

  fprintf(stderr, "upload baud = %s\n", uBaudStr);
    
  if (!strcmp(nBaudStr, "9600")) {
    nBaud = B9600;
  }
  else if (!strcmp(nBaudStr, "19200")) {
    nBaud = B19200;
  }
  else if (!strcmp(nBaudStr, "38400")) {
    nBaud = B38400;
  }
#ifdef B57600
  else if (!strcmp(nBaudStr, "57600")) {
    nBaud = B57600;
  }
#endif
#ifdef B115200
  else if (!strcmp(nBaudStr, "115200")) {
    nBaud = B115200;
  }
#endif
#ifdef B230400
  else if (!strcmp(nBaudStr, "230400")) {
    nBaud = B230400;
  }
#endif
  else {
    fprintf(stderr, "Unknown normal baud, using 38400");
    strcpy(nBaudStr, "38400");
    nBaud = B38400;
  }

  if (!strcmp(uBaudStr, "9600")) {
    uBaud = B9600;
  }
  else if (!strcmp(uBaudStr, "19200")) {
    uBaud = B19200;
  }
  else if (!strcmp(uBaudStr, "38400")) {
    uBaud = B38400;
  }
#ifdef B57600
  else if (!strcmp(uBaudStr, "57600")) {
    uBaud = B57600;
  }
#endif
#ifdef B115200
  else if (!strcmp(uBaudStr, "115200")) {
    uBaud = B115200;
  }
#endif
#ifdef B230400
  else if (!strcmp(uBaudStr, "230400")) {
    uBaud = B230400;
  }
#endif
  else {
    fprintf(stderr, "Unknown upload baud, using 38400");
    strcpy(uBaudStr, "38400");
    uBaud = B38400;
  }

  /* open and configure serial port */
  fd = openRaw(port, O_RDWR | O_NONBLOCK);
  if (fd<0) {
    fprintf(stderr, "unable to open port\n");
    exit(1);
  }

  if (!setBaudRate(fd, nBaud)) {
    fprintf(stderr, "unable to set baud rate on port\n");
    exit(1);
  }

  reflex = reflexNew(fd, NULL);

  reflex->handler                 = defaultHandlePort;

  reflex->system->lcdDumpRpl      = sysLcdDumpRpl;
  reflex->system->statusRpl       = sysStatusRpl;
  reflex->system->unknownRpl      = sysUnknownRpl;

  reflex->motion->getSystemRpl    = motAxisGetSystemRpl;
  reflex->motion->getModelRpl     = motAxisGetModelRpl;
  reflex->motion->getTargetRpl    = motAxisGetTargetRpl;
  reflex->motion->setDirRpl       = motAxisSetDirRpl;
  reflex->motion->setDefaultsRpl  = motSetDefaultsRpl;
  reflex->motion->brakeSetRpl     = motBrakeSetRpl;
  reflex->motion->brakeReleaseRpl = motBrakeReleaseRpl;
  reflex->motion->setSystemRptRpl = motSetSystemRptRpl;
  reflex->motion->unknownRpl      = motUnknownRpl;
  reflex->motion->systemRpt       = motAxisSystemRpt;

  reflex->joystick->stateRpl      = jstkGetStateRpl;
  reflex->joystick->unknownRpl    = jstkUnknownRpl;

  reflex->sonar->updateRpl        = sonarGetUpdateRpl;
  reflex->sonar->rptRpl           = sonarRptRpl;
  reflex->sonar->unknownRpl       = sonarUnknownRpl;
  reflex->sonar->rpt              = sonarRpt;

  dioTS = 0;
  dioCount = 0;
  dioOn = 0;

  reflex->dio->updateRpl          = dioUpdateRpl;
  reflex->dio->update             = dioUpdate;
  reflex->dio->rptRpl             = dioRptRpl;
  reflex->dio->rpt                = dioRpt;
  reflex->dio->unknownRpl         = dioUnknownRpl;

  /*
   * Catch exit signals so that we can reset the tty mode.
   */
 
  if (signal(SIGINT, sig_catch) == SIG_ERR) {
    fprintf (stderr, "signal(SIGINT) error\n");
    exit (0);
  }
  if (signal(SIGQUIT, sig_catch) == SIG_ERR) {
    fprintf(stderr, "signal(SIGQUIT) error\n");
    exit (0);
  }
  if (signal(SIGTERM, sig_catch) == SIG_ERR) {
    fprintf(stderr, "signal(SIGTERM) error\n");
    exit (0);
  }

  /*
   * Set new tty mode
   */

  if (tty_cbreak(STDIN_FILENO) < 0) {
    fprintf(stderr, "tty_cbreak() error\n");
    exit (0);
  }

  readLoop(reflex, fd);
  return 1;
}
