/*
 * Copyright (c) 2002-2005 CNRS/LAAS
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/* $Id: serial.c,v 1.2 2005/08/22 09:17:47 matthieu Exp $ */

#include <unistd.h>
#include <stdlib.h> 
#include <stdio.h> 
#include <string.h>
#include <termios.h>
#include <sys/time.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <signal.h>
#include <ctype.h>

#include "serial.h"

/**********************************************************/
/*                                                        */
/*  openRaw                                               */    
/*                                                        */
/*  Opens passed filename and sets its device to raw mode.*/
/*  Returns file descriptor.                              */
/*                                                        */
/*  raw mode is defined as:                               */
/*  Off: echo, canonical input, extended processing,      */
/*       signals, break key, parity, 8th bit strip,       */
/*       flow control, output post processing             */
/*   On: 8 bit size                                       */
/*                                                        */
/**********************************************************/
 
int 
openRaw(const char * filename, mode_t io_flags)
{
	struct termios term_info;
	int fd;
	
	fd = open(filename,io_flags);
	if (fd == -1) {
		fprintf(stderr, "ERROR: failed to open %s\n", filename);
		return -1;
	}
	
	if(tcgetattr(fd,&term_info) <0) {
		fprintf(stderr, "ERROR: %s is not a terminal\n", filename);
		close(fd);
		return -1;
	}
	
	/* turn off echo, canonical mode, extended processing, signals */
	term_info.c_lflag &= ~(ECHO | ICANON | IEXTEN | ISIG);
	
	/* turn off break sig, cr->nl, parity off, 8 bit strip, flow control */
	term_info.c_iflag &= ~(BRKINT | ICRNL | INPCK | ISTRIP | IXON);
	
	/* clear size, turn off parity bit */
	term_info.c_cflag &= ~(CSIZE | PARENB);
	
	/* set size to 8 bits */
	term_info.c_cflag |= CS8;
	
	/* turn output processing off */
	term_info.c_oflag &= ~(OPOST);
	
	/* Set time and bytes to read at once */
	term_info.c_cc[VTIME] = 0;
	term_info.c_cc[VMIN] = 0;
	
	if(tcsetattr(fd,TCSAFLUSH,&term_info) <0)
	{
		fprintf(stderr, "ERROR: failed to set control attributes to %s\n", filename);
		close(fd);
		return -1;
	}
	
	return fd;
}

static struct speed {
	int speed;
	speed_t termioSpeed;
} speeds[] = {
	{ 0, B0 },
	{ 50, B50 },
	{ 75, B75 },
	{ 110, B110 },
	{ 300, B300 },
	{ 600, B600 },
	{ 1200, B1200 },
	{ 2400, B2400 },
	{ 4800, B4800 },
	{ 9600, B9600 },
	{ 19200, B19200 },
	{ 38400, B38400 },
#ifdef B57600
	{ 57600, B57600 },
#endif
#ifdef B115200
	{ 115200, B115200 }
#endif
};

#define SPEED_SIZE (sizeof(speeds)/sizeof(struct speed))

static speed_t
get_termios_speed(int speed) 
{
	int i;
	
	for (i = 0; i < SPEED_SIZE; i++) {
		if (speed == speeds[i].speed) {
			return speeds[i].termioSpeed;
		}
	} /* for */
	return -1;
}

/**********************************************************/
/*                                                        */
/*  setBaudRate                                           */    
/*                                                        */
/*  Set the baud rate of an open device.                  */
/*                                                        */
/*  Baud rate constants are found in <termios.h>.         */
/*  B9600 == 9600 baud                                    */
/*                                                        */
/**********************************************************/

int 
setBaudRate (int fd, int baud)
{
	struct termios terminfo;
	speed_t baudRate;
	int error;
	
	baudRate = get_termios_speed(baud);
	if (baudRate == -1) {
		return FALSE;
	}
	
	error = tcgetattr(fd, &terminfo);
	if (error) {
		perror("tcgetattr()");
		return(FALSE);
	}
	
	error = cfsetospeed(&terminfo, baudRate);
	if (error) {
		fprintf(stderr, "cfsetospeed(%ld): ", (long)baudRate);
		perror(NULL);
		return(FALSE);
	}
	
	error = cfsetispeed(&terminfo, baudRate);
	if (error) {
		fprintf(stderr, "cfsetispeed(%ld): ", (long)baudRate);
		perror(NULL);
		return(FALSE);
	}
	
	error = tcsetattr(fd, TCSANOW, &terminfo);
	if (error) {
		perror("tcsetattr()");
		return(FALSE);
	}
	
	return(TRUE);
}

